#include "HierClustering.h"
#include <assert.h>
#include <algorithm>
#include <iterator>
#include "Debug.h"

#include "mex.h" // just for warnings etc.

IntUPairList uniqueIntUPairs(const IntUPairList &input)
{
	IntUPairList sortInput = input, unique;
	if (input.empty()) return unique;
	std::sort(sortInput.begin(), sortInput.end());
	
	IntUPair currEdge = sortInput[0];
	for (int e=1; e<(int)sortInput.size(); e++)
	{
		if (!(sortInput[e] == currEdge))
		{
			unique.push_back(currEdge);
			currEdge = sortInput[e];
		}
	}
	unique.push_back(currEdge);

	return unique;
}

void HierClusteringSlow::init(int numNodes, double *pValues, double *pWeights, int numEdges, double *pEdges)
{
	m_active.assign(numNodes, true);
	m_nodeValues.resize(numNodes);
	m_nodeWeights.resize(numNodes);
	m_clusterElements.resize(numNodes);
	for (int n=0; n<numNodes; n++)
	{
		m_nodeValues[n] = pValues[n];
		m_nodeWeights[n] = pWeights[n];
		m_clusterElements[n].insert(n);

		if (m_nodeWeights[n] < 1e-7) mexWarnMsgTxt("Very small triangle area");				
	}

	IntUPairList inpEdges;
	inpEdges.resize(numEdges);
	bool zeroTest = false;
	for (int e=0; e<numEdges; e++)
	{
		int v1 = (int)pEdges[e] - 1;
		int v2 = (int)pEdges[e + numEdges] - 1;
		assert(v1 >= 0 && v2 >= 0);
		if (v1 == 0 || v2 == 0) zeroTest = true;
		inpEdges[e] = IntUPair(v1, v2);
	}
	assert(zeroTest); // otherwise we probably incorrectly converted from MATLAB indexing -- just a test :)

	// kill duplicates:
	m_edges = uniqueIntUPairs(inpEdges);
}

double HierClusteringSlow::clusterMean(const IntSet &nodes)
{
	double sumWeightedValues = 0.0;
	double sumWeights = 0.0;

	for (IntSet::const_iterator it = nodes.begin(); it != nodes.end(); it++)
	{
		int nodeIdx = *it;
		assert(nodeIdx >= 0 && nodeIdx < getNumNodes());
		sumWeightedValues += m_nodeValues[nodeIdx] * m_nodeWeights[nodeIdx];
		sumWeights += m_nodeWeights[nodeIdx];
	}

	return sumWeightedValues / sumWeights;
}

double HierClusteringSlow::clusterVar(const IntSet &nodes)
{
	const double cMean = clusterMean(nodes);
	double sumVar = 0.0;

	for (IntSet::const_iterator it = nodes.begin(); it != nodes.end(); it++)
	{
		int nodeIdx = *it;
		assert(nodeIdx >= 0 && nodeIdx < getNumNodes());

		double x = (m_nodeValues[nodeIdx] - cMean);
		sumVar += x * x * m_nodeWeights[nodeIdx];
	}

	return sumVar;
}

int HierClusteringSlow::findBestMergeEdge(double &bestMergeCost)
{
	bestMergeCost = 1e10;
	int bestEdge = -1;

	for (int e=0; e<getNumEdges(); e++)
	{
		const int v1 = m_edges[e].getV1();
		const int v2 = m_edges[e].getV2();

		assert(m_active[v1] && m_active[v2]);		
		const double cost1 = clusterVar(m_clusterElements[v1]);
		const double cost2 = clusterVar(m_clusterElements[v2]);

		IntSet mergedNodes; 
		std::set_union(m_clusterElements[v1].begin(), m_clusterElements[v1].end(), m_clusterElements[v2].begin(), m_clusterElements[v2].end(), std::inserter(mergedNodes, mergedNodes.end()));
		const double cost12 = clusterVar(mergedNodes);

		double mergeCost = cost12 - cost1 - cost2;
		if (mergeCost < -1e-10)
		{
			TRACE("Warning: mergeCost = %.20f\n", mergeCost);
		}
		if (mergeCost < bestMergeCost)
		{
			bestMergeCost = mergeCost;
			bestEdge = e;
		}
	}
	
	return bestEdge;
}

void HierClusteringSlow::clustering(int targetNumClusters)
{
	const int numSteps = getNumNodes() - targetNumClusters;

	for (int step=0; step<numSteps; step++)
	{
		// find which two clusters to merge:
		double bestMergeCost;
		int bestEdge = findBestMergeEdge(bestMergeCost);		

		// merge clusters -- cluster v2 merged into v1:
		const int v1 = m_edges[bestEdge].getV1();
		const int v2 = m_edges[bestEdge].getV2();
		TRACE("step %i: merging clusters %i and %i at cost %.15f\n", step, v1, v2, bestMergeCost);

		m_active[v2] = false;
		m_clusterElements[v1].insert(m_clusterElements[v2].begin(), m_clusterElements[v2].end());
		m_clusterElements[v2].clear();

		// replace v2 by v1 in m_edges:
		IntUPairList repEdges;
		for (int e=0; e<getNumEdges(); e++)
		{
			int ev1 = m_edges[e].getV1();
			int ev2 = m_edges[e].getV2();
			m_edges[e].assertOrder();

			if (ev1 == v2) {	
				if (v1 != ev2) repEdges.push_back(IntUPair(v1, ev2));				
			} 
			else if (ev2 == v2)
			{
				if (v1 != ev1) repEdges.push_back(IntUPair(v1, ev1));								
			}
			else
			{
				repEdges.push_back(m_edges[e]);
			}
		}

		m_edges = uniqueIntUPairs(repEdges);

		//// this could have created duplicates, so remove them:
		//std::sort(repEdges.begin(), repEdges.end());
		//IntUPairList newEdges;
		//IntUPair currEdge = repEdges[0];
		//for (int e=1; e<(int)repEdges.size(); e++)
		//{
		//	if (!(repEdges[e] == currEdge))
		//	{
		//		newEdges.push_back(currEdge);
		//		currEdge = repEdges[e];
		//	}
		//}
		//newEdges.push_back(currEdge);
	
		//m_edges = newEdges; // duplicates removed

		int hu = 1;
	}
}

void HierClusteringSlow::writeBackResult(double *pRes)
{
	for (int n=0; n<getNumNodes(); n++) pRes[n] = -1.0;

	int renumber = 1;
	for (int i=0; i<(int)m_clusterElements.size(); i++)
	{
		bool active = false;
		for (IntSet::iterator it = m_clusterElements[i].begin(); it != m_clusterElements[i].end(); it++)
		{
			int el = *it;
			pRes[el] = renumber;
			//pRes[el] = i + 1; // convert to MATLAB indexing
			active = true;
		}
		if (active) renumber++;
	}
}