#ifndef CONNECTIVITY_H
#define CONNECTIVITY_H

#include "Tuples.h"
#include "types.h"
#include <map>
#include <queue>

// for d = 2, elements are triangles and facets are edges
// for d = 3, elements are tetrahedra and facets are triangles
template <unsigned int d> class Peeling 
{
public:	
	typedef std::map<Tuple<d>, Tuple<2> > FacetElementMap;
	typedef std::map<Tuple<2>, std::vector<int> > EdgeElementMap;
	typedef std::map<int, std::vector<int> > VertexElementMap;
	typedef std::set<Tuple<d> > FacetSet;

	Peeling() {}

	void init(int numElements, const double *pElements, const double *pValues, int numFacets, const double *pFacets);
	void init(int numElements, const double *pElements, const double *pValues, int numFacets, const double *pFacets,
            const std::map<std::pair<int,int>,double> & pSmoothnessTerm,
            const std::vector<double> & pDataTermIn,
            const std::vector<double> & pDataTermOut
            );
	void peel(BoolVector &outside);
	// more careful peeling version which dutifully searches for smallest weight element that can be peeled out
	// if 'peelFromOutside' is true, then only elements that are adjacent to the boundary or another element outside can be peeled out:
        //
        // Inputs:
        //   outside  empty list of bools revealing whose outside and inside
        //   peelFromOutside  only peel away from boundary or already peeled
        //     elements
        //   singleCC  maintain a single connected component
        //   minCutCapacity  peel based on improvement of the graph cut
        //     energy
        // Outputs:
        //   outside  list of bools revealing whose outside and inside
	void peelOneByOne(BoolVector &outside, 
          const bool peelFromOutside,
          const bool singleCC = true,
          const bool minCutCapacity = false);
        // Improve an existing segmentation while maintaining constraints
        //
        // Inputs:
        //   outside  current list of bools revealing whose outside and inside
        //   dirtyQ  a queue of elements which are "dirty"
        //   must_be_legal  force legal moves
        // Outputs:
        //   outside  list of bools revealing whose outside and inside
        // Returns the number of flips
        int improve(
          BoolVector &outside, 
          std::queue<int> & dirtyQ, 
          const bool must_be_legal);
        // Project existing segmentation to meet constraints
        //
        // Inputs:
        //   outside  current list of bools revealing whose outside and inside
        // Outputs:
        //   outside  list of bools revealing whose outside and inside
        void project(BoolVector &outside);
        // Project facet constraints
        void project_constrained_facets(BoolVector &outside);
        // Project manifold-edge constraints
        void project_edge_manifold(BoolVector &outside);
        // Project manifold-vertex constraints
        void project_vertex_manifold(BoolVector &outside);
        // Given a list of elements (around some edge or vertex) add enough
        // them to the inside so that the dual graph of the inside elements is
        // a single connected component
        //
        // Inputs:
        //   Ele  list of element indices
        //   outside  current list of bools revealing whose outside and inside
        // Outputs:
        //   outside  list of bools revealing whose outside and inside
        void project_single_cc(
            const std::vector<int> & Ele,
            BoolVector &outside);
        // ... using greedy-blind-dumb method
        void project_single_cc_blind(
            const std::vector<int> & Ele,
            BoolVector &outside);
        // ... using BFS method
        void project_single_cc_bfs(
            const std::vector<int> & Ele,
            BoolVector &outside);
        //// Compute the number of connected components in a given list of
        //// elements
        ////
        //// Inputs:
        ////   Ele  list of element indices
        //// Returns number of connected components

        //// Compute the cost of setting an element in the current list of outside
        //// elements to out
        ////
        //// Inputs:
        ////   outside  current list of bools revealing whose outside and inside
        ////   e  index of element to consider
        ////   out  whether measuring cost if e is out or in
        //// Returns cost of treating e as out
        //double cost(const BoolVector & outside, const int e, const bool out);
        // Total cut-capacity (cost)
        //
        // Inputs:
        //   outside  current list of bools revealing whose outside and inside
        // Returns total cost
        double cut_capacity(const BoolVector & outside);
        // Compute change in cut capacity if outside[e] is set to out
        //
        // Inputs:
        //   outside  current list of bools revealing whose outside and inside
        //   e  index of element to consider
        //   out  whether measuring cost if e is out or in
        // Return the change in cut capacity 
        double cut_capacity_change(const BoolVector & outside, const int e, const bool out);
        // Compute whether it would be legal to set outside[e] to out 
        //
        // Inputs:
        //   outside  current list of bools revealing whose outside and inside
        //   e  index of element to consider
        //   out  whether measuring cost if e is out or in
        // Return the change in cut capacity 
        bool is_legal(const BoolVector & outside, const int e, const bool out);

        // Find all elements incident on (undirected) edge {u,v}
        //
        // Inputs:
        //   u  source of edge
        //   v  destination of edge
        // Returns list of indices of elements incident on {u,v}
        std::vector<int> incident(const int u, const int v);
        // Same as above but incident on vertex i
        std::vector<int> incident(const int i);

        // Determine if the dual graph of the given list of indices forms a
        // single connected component
        //
        // Inputs:
        //   Ele  list of element indices
        // Returns true only if the number of components in the dual graph is
        // exactly one
        bool single_cc(const std::vector<int> & Ele);

	// creates m_facetElementMap:
	bool prepareFacetElementMap();
	void prepareEdgeElementMap();
	void prepareVertexElementMap();
	// creates m_constrFacetSet:
	void prepareConstrFacetSet();
	// creates m_elemNeighbors and m_constrNeighbors:
	void prepareElemNeighbors();	

	// checks whether each facet specified in m_constrFaces is a facet of some element in m_elements:
	bool testConstrFacetsPresence();

	// returns elements sorted by values:
	DoubleIntPairVector sortElementsByValues();

	inline int getNumElements() const { return (int)m_elements.size(); }

	// returns size of connected component of elements that have 'outside' flag false, starting from 'start':
	int connectedComponentInside(const BoolVector &outside, int start);

protected:
	// original input data:	
	TupleVector<d+1> m_elements; // list of elements with original ordering
	DoubleVector m_values; // per-element values
	TupleVector<d> m_constrFacets; // list of constrained facets with original ordering

	// maps each ordered facet to indices of its adjacent elements (at most 2), -1 if none:
	FacetElementMap m_facetElementMap;
        // map each ordered edge to incident elements
	EdgeElementMap m_edgeElementMap;
        // map each vertex to incident elements
	VertexElementMap m_vertexElementMap;

	// set of ordered constrained facets:
	FacetSet m_constrFacetSet;

	// neighboring elements (connectivity):
	TupleVector<d+1> m_elemNeighbors; // size getNumElements, stores indices into m_elements of neighboring elements or -1 if none
	TupleVector<d+1> m_elemConstrNeighbors; // same as m_elemNeighbors, but stores flag whether the neighbor is connected by a constrained facet (1) or not (0)
  // Graphcut energy parameters

  // A better way to handle this would be to just load in the exact
  // graphcut graph and use that directly
  std::map<std::pair<int,int>,double> m_SmoothnessTerm;
  std::vector<double> m_DataTermIn;
  std::vector<double> m_DataTermOut;


  // Value at source and dest (defines data energy term)
  double source_v;
  double dest_v;
  // Smoothness weight
  //   larger --> less splintery, but less faithful to data
  //   smaller --> more faithful to local data, more splintery
  double smooth_w;
  // Data term weight. This is technically not necessary, since all the
  // balancing can be done with the smoothness weight. BUT this is used to
  // coincide with the matlab graphcut implementation which needs integer
  // weights
  double data_w;

  // Non-manifoldness

};

#endif
